/*=============================================================================
	PartialHeader.h:
	This is not supposed to build and should only be used for learning purposes
=============================================================================*/

#if _MSC_VER
#pragma pack (push,4)
#endif

#ifndef XC_ENGINE_API
#define XC_ENGINE_API DLL_IMPORT
#endif

#ifndef NAMES_ONLY
#define AUTOGENERATE_NAME(name) extern XC_ENGINE_API FName XC_ENGINE_##name;
#define AUTOGENERATE_FUNCTION(cls,idx,name)
#endif

AUTOGENERATE_NAME(XC_Engine)
AUTOGENERATE_NAME(XC_GameEngine)
AUTOGENERATE_NAME(XC_ConnectionHandler)
AUTOGENERATE_NAME(PreLoginHook)
AUTOGENERATE_NAME(AdminLoginHook)

#ifndef NAMES_ONLY

//Do remove debugging instructions
//#define CACUS_DEBUG

#define MAX_OCTREE_DEPTH 6
#define OCTREE_GRID_ARRAY 7
#define TWO_EXP_OGA 64 //2^OCTREE_GRID_ARRAY

#define MIN_ACTORS_FOR_SUBNODE 3


//These variables should persist, so we can't really afford to let the garbage collector to mess this up
extern UXC_GameEngine*	GEngineXC; //XC_GameEngine object, for use with soft-hook mode
XC_ENGINE_API extern FCollisionCacus*	CollisionCacus;
XC_ENGINE_API extern FXC_TravelManager*	TravelManager;
XC_ENGINE_API extern FXC_TimeManager*	TimeManager;
XC_ENGINE_API extern INT				b451Setup;
XC_ENGINE_API extern INT				TmpSize;
XC_ENGINE_API extern UBOOL				CanEditPackageMap;
XC_ENGINE_API extern UBOOL				LogNet;
XC_ENGINE_API extern INT				ExtraThreads;

XC_ENGINE_API extern INT				OctreeGridAssignment[TWO_EXP_OGA];

XC_ENGINE_API extern INT				DebugCycles;
XC_ENGINE_API extern INT				RayCycles;
XC_ENGINE_API extern INT				TraceCycles;
XC_ENGINE_API extern INT				TotalCycles;
XC_ENGINE_API extern INT				GridCycles;

/*
static void TestInverse()
{
	for ( FLOAT f=0 ; f<33000 ; f = f*1.4f + 0.005f )
	{
		FLOAT inv = _Reciprocal(f);
		FLOAT div = 1.f/f;
		debugf( NAME_Log, TEXT("Test inverse of %f: %f, %f, %f"), f, div, inv, div-inv);
	}
}
*/

class FPrecomputedRay
{
public:
	FVector Org;
	FVector End;
	FVector Dir;
	FVector Inv;
	FBox ExtendedBox; //For grid interaction and quick rejects
	FVector XAxis; //For CylCyl checks
	FVector YAxis;
	const FVector* Extent;
	FLOAT Dist;
	UBOOL(FPrecomputedRay::*TraceFunc)( AActor*, FVector*, FVector*);

	FPrecomputedRay()	{};
	FPrecomputedRay( const FVector& A, const FVector& B, const FVector& TExtent);


	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X <= Container.Max.X) && (Sample.Y <= Container.Max.Y) && (Sample.Z <= Container.Max.Z);
	}

	void TwoPointBox( FBox& Box, const FVector& A, const FVector& B);
	INT IntersectsBox( const FBox& Box);

	UBOOL HitsGCylActor( AActor* Actor, FVector* HitLocation, FVector* HitNormal);
	UBOOL HitsVCylActor( AActor* Actor, FVector* HitLocation, FVector* HitNormal);
	UBOOL HitsCylActor( AActor* Actor, FVector* HitLocation, FVector* HitNormal);
};


struct XC_ENGINE_API FCacusActorInfo
{
	AActor* Actor;
	INT SubOctant; //-1 if not
	FBox Box;
	FVector Location;
	FCacusActorInfo( AActor* NewActor, INT NewOctant=-1)
	:	Actor(NewActor)
	,	SubOctant(NewOctant)
	,	Box( NewActor->GetPrimitive()->GetCollisionBoundingBox(NewActor) )
	,	Location( NewActor->Brush ? ((Box.Min+Box.Max)*0.5) : NewActor->Location)
	{
		if ( NewActor->Brush )
			Box.ExpandBy( 1.f);
	}
};

class XC_ENGINE_API FCollisionOctree
{
public:
	FCollisionCacus* Master;
	FCollisionOctree* Parent;
	FCollisionOctree* Children[8]; //X,Y,Z based bitwise.
	FCollisionOctree* ReducedChildren[8]; //For fast operations
	FBox BoundingBox;
	FVector CenterPoint;
	BYTE CurDepth;
	BYTE ChildIndex;
	BYTE ChildCount;
	BYTE CleanupTimer;
	BYTE XPos, YPos, ZPos;
	INT MIndex; //Index on master array
	TArray<FCacusActorInfo> Actors;
	TArray<FCacusActorInfo> BrushMovers; //Brush && bIsMover
	//FCollisionCacus uses this
	FCollisionOctree()
	:	CurDepth(0)
	,	ChildIndex(0)
	,	ChildCount(0)
	,	CleanupTimer(20)
	,	Actors(0)
	,	BrushMovers(0)
	,	CenterPoint(0,0,0)
	,	XPos(0)	,	YPos(0)	,	ZPos(0)
	,	BoundingBox( FVector(-32768,-32768,-32768), FVector(32768,32768,32768))
	{
		appMemzero( &Parent, 4*17);
	}
	FCollisionOctree( FCollisionOctree* InParent, FBox InBounds, INT InDepth, BYTE InChild=0)
	:	Parent(InParent)
	,	BoundingBox(InBounds)
	,	CenterPoint( (InBounds.Min+InBounds.Max) * 0.5)
	,	CurDepth(InDepth)
	,	ChildIndex( InChild)
	,	CleanupTimer(20)
	,	ChildCount(0)
	,	XPos(0)	,	YPos(0)	,	ZPos(0)
	,	Actors(0)
	,	BrushMovers(0)
	{
		appMemzero( &Children[0], 4*16 ); //Both child arrays
		if ( InParent )
		{
#ifdef CACUS_DEBUG
			check( InParent->ChildCount < 8 );
			check( !InParent->ReducedChildren[InParent->ChildCount] );
#endif
			InParent->ReducedChildren[InParent->ChildCount] = this;
			InParent->ChildCount++;
		}
#ifdef CACUS_DEBUG
		check( InBounds.Min.X < InBounds.Max.X );
		check( InBounds.Min.Y < InBounds.Max.Y );
		check( InBounds.Min.Z < InBounds.Max.Z );
//		debugf( NAME_Log, TEXT("[XC_GE] Creating new Octree node, Depth %i"), CurDepth);
#endif
	}
	~FCollisionOctree();

	void Cleanup();
	FBox GetSubOctantBox( const FVector& Point, INT& Idx);		//Fast, does not do sanity checks
	INT GetSubOctantIdx( const FVector& Point);				//Same
	FCacusActorInfo* FindActor( FName ByName);
	FIteratorActorList* RadiusQuery( FMemStack& Mem, const FVector& Location, const FLOAT& Radius, const DWORD& ExtraNodeFlags, FIteratorActorList* PrevLink );
	FCheckResult* PointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink );
	FCheckResult* FastPointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink);
	FCheckResult* MEncroachQuery( FMemStack& Mem, AActor* Actor, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* PEncroachQuery( FMemStack& Mem, AActor* Actor, const FVector& Location, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* AEncroachQuery( FMemStack& Mem, AActor* Actor, const FVector& Location, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* LineQuery
	(
	 FMemStack& Mem
	, const FVector& End
	, const FVector& Start
	, const FVector& Extent
	, const BYTE& ExtraNodeFlags
	, FPrecomputedRay& Ray
	,  FCheckResult* PrevLink
	);
	FCheckResult* FastLineQuery
	(
	 FMemStack& Mem
	, const FVector& End
	, const FVector& Start
	, const FVector& Extent
	, const BYTE& ExtraNodeFlags
	, FPrecomputedRay& Ray
	, FCheckResult* PrevLink
	);
	inline UBOOL FitsBox( const FBox& Sample, const FBox& Container)
	{
		return (Sample.Min.X >= Container.Min.X) && (Sample.Min.Y >= Container.Min.Y) && (Sample.Min.Z >= Container.Min.Z)
			&& (Sample.Max.X <= Container.Max.X) && (Sample.Max.Y <= Container.Max.Y) && (Sample.Max.Z <= Container.Max.Z);
	}
	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X < Container.Max.X) && (Sample.Y < Container.Max.Y) && (Sample.Z < Container.Max.Z);
	}
	inline UBOOL FitsBox( AActor* Actor, const FBox& Container)
	{
		return (Actor->Location.X - Actor->CollisionRadius >= Container.Min.X) && (Actor->Location.Y - Actor->CollisionRadius >= Container.Min.Y) && (Actor->Location.Z - Actor->CollisionHeight >= Container.Min.Z)
			&& (Actor->Location.X + Actor->CollisionRadius < Container.Max.X) && (Actor->Location.Y + Actor->CollisionRadius < Container.Max.Y) && (Actor->Location.Z + Actor->CollisionHeight < Container.Max.Z);
	}
	inline FBox ExpandByExtent( const FBox& Box, const FVector& Extent)
	{	return FBox( Box.Min-Extent, Box.Max+Extent);	}
	//Quick reject, actor should not go into a sub-box
	inline UBOOL HitsCenterAxes( const FBox& Box)
	{
		return ( ((Box.Min.X < CenterPoint.X) && (Box.Max.X > CenterPoint.X))
		||	((Box.Min.Y < CenterPoint.Y) && (Box.Max.Y > CenterPoint.Y))
		||	((Box.Min.Z < CenterPoint.Z) && (Box.Max.Z > CenterPoint.Z)) );
	}
};

//====================================
// Three dimension grid
//==========
class XC_ENGINE_API FNodeGrid3 : protected TArray<FCollisionOctree*>
{
public:
	INT GridSize;
//	INT Sub[3]; //FPU is causing lots of precision issues
	FVector Substract; //Substract this when calculating grid position, should be identical in all grids?
	FVector BoxSize; //Dividing a vector by this (post substraction) will give coordinates!
	FVector Divider; //Precalculated version

	FNodeGrid3()
	:	TArray<FCollisionOctree*>()
	,	GridSize(0)
	{};
	inline INT Num() const
	{
		return ArrayNum;
	}
	void Empty( INT Slack=0 )
	{
		TArray<FCollisionOctree*>::Empty( Slack );
	}
	void SetBaseGrid( FCollisionOctree* Base);
	void SetGrid( INT NewGridSize, const FBox& Bounds);
	void Set( FCollisionOctree* In);
	void Remove( FCollisionOctree* In);
	FCollisionOctree* Get( FVector Pos);
};


class XC_ENGINE_API FCollisionCacus : public FCollisionHashBase , public FCollisionOctree
{
public:
	FCollisionHashBase* Hooked;
	INT MinAvailable;
	TArray<FCollisionOctree*> ColList;
	AActor* JustRemoved;
	FNodeGrid3 DepthGrid[OCTREE_GRID_ARRAY]; //Zero isn't used, we're doing this for quick CurDepth access

	FCollisionCacus()
	:	ColList(0)
	,	MinAvailable(0)
	{
		ColList.AddItem( this);
		MIndex = 0;
		Master = this;
		appMemzero( &DepthGrid, sizeof(class FNodeGrid3) * OCTREE_GRID_ARRAY );
	};
	~FCollisionCacus();

	void SetupGrids();
	void ChildSet( FCollisionOctree* Child);
	void Tick();
	void AddActor( AActor *Actor );
	void RemoveActor( AActor *Actor );
	FCheckResult* ActorLineCheck( FMemStack& Mem, FVector End, FVector Start, FVector Extent, BYTE ExtraNodeFlags );
	FCheckResult* ActorPointCheck( FMemStack& Mem, FVector Location, FVector Extent, DWORD ExtraNodeFlags );
	FCheckResult* ActorRadiusCheck( FMemStack& Mem, FVector Location, FLOAT Radius, DWORD ExtraNodeFlags );
	FCheckResult* ActorEncroachmentCheck( FMemStack& Mem, AActor* Actor, FVector Location, FRotator Rotation, DWORD ExtraNodeFlags );
	void CheckActorNotReferenced( AActor* Actor );
};


struct UHack_eventPreLoginHook_Parms
{
    FString Options;
    FString Address;
    FString Error;
    FString FailCode;
};

struct UHack_eventAdminLoginHook_Parms
{
	class APlayerPawn* P;
	BITFIELD ReturnValue;
};

struct MappedPropertyBlock
{
	INT Idx; //Position in array
	INT ParentIdx;
	INT TotalProps; //Faster accessing
	UClass* Class; //Actor class
	TArray< UProperty* > Properties;
	TArray< INT > ChildrenIdx;

	~MappedPropertyBlock()
	{
		if ( Properties.Num() > 0 )
			Properties.Empty();
		if ( ChildrenIdx.Num() > 0 )
			ChildrenIdx.Empty();
	}
};

//Used when for inventory/pawn travel
class XC_ENGINE_API FTravelClassMap : public TArray< MappedPropertyBlock >
{
	public:
	INT RegisterClass( UClass* InClass); //Return index in map
};

//Used for transient Actor reference cleanup
class XC_ENGINE_API FTransientClassMap : public TArray< MappedPropertyBlock >
{
	public:
	FTransientClassMap( INT i=0)
	:	TArray< MappedPropertyBlock>(0)
	{};
	INT RegisterClass( UClass* InClass);
};

struct FTravelPlayerMap
{
	INT PlayerID;
	FString PlayerName;
	FString TravelList;
	FTravelPlayerMap() {};
};

class XC_ENGINE_API FXC_TravelManager : public FExec
{
	public:
	UXC_GameEngine*	Engine;
	BITFIELD		bAutoMode:1 GCC_PACK(4); //Do stuff myself
	BITFIELD		bNoPoll:1; //Do not operate during this level
	FLOAT			TimerCounter GCC_PACK(4); //For automatic manager

	TArray<FTravelPlayerMap> PlayerMap;
	FTravelClassMap	TravelClassMap; //Do not serialize, we want FLUSH and mapswitches to clear this
	
	FXC_TravelManager( UXC_GameEngine* InEngine=NULL)
	:	Engine(InEngine)
	,	TimerCounter( 2.f)
	{};
	
	//FExec interface
	UBOOL Exec( const TCHAR* Cmd, FOutputDevice& Ar );
	
	//FXC_TravelManager interface
	void Poll( FLOAT DeltaTime, UBOOL bForceUpdate=0);
	void FixUpPlayerId( APlayerPawn* P); //Checks for name changes
	void GenerateTravelInfo( APlayerPawn* P);
//	void PreMapChange( ULevel* OldLevel);
	void PostMapChange( ULevel* NewLevel, UBOOL bClearMap=0);
	
};


class AXC_DummyPawn : private APawn
{
	friend class UXC_GameEngine;
	
	public:
	DECLARE_FUNCTION(NewMapName);			//539 - hack
	DECLARE_FUNCTION(NewPlayerCanSeeMe);	//532 - hack
	DECLARE_FUNCTION(NewDynArrayElement);	//10 - version 10
	DECLARE_FUNCTION(NewDynArrayLength);	//640 - version 10
	DECLARE_FUNCTION(NewDynArrayInsert);	//641 - version 10
	DECLARE_FUNCTION(NewDynArrayRemove);	//642 - version 10
	DECLARE_FUNCTION(NewAppSeconds);		//643 - version 10
	DECLARE_FUNCTION(NewMakeColor);			//192 - version 10
	DECLARE_FUNCTION(execAddToPackagesMap);	//1718 - version 11
    DECLARE_FUNCTION(PreLoginNative);		//3550 - internal
    DECLARE_FUNCTION(AdminLoginNative);		//3551 - internal
    DECLARE_FUNCTION(execCollidingActors);	//3552 - version 12
    DECLARE_FUNCTION(execIsInPackageMap);	//1719 - version 13
	//TODO:
	//native(601) final function Class<Object> GetParentClass( Class<Object> ObjClass );
	/* Lower-case some text */
	//static native(238) final function string Locs( string InStr );
	/*
	native(1715) final function Actor SpawnAct
	(
		Class<Actor>	SpawnClass,
		vector		SpawnLocation,
		optional rotator	SpawnRotation,
		optional name	ActorName,
		optional Actor	SpawnOwner,
		optional Pawn	SpawnInstigator,
		optional Actor	Template,
		optional bool	bMayColFail
	);
	*/
	protected:
	UBOOL PCanSee( APlayerPawn* Other);

//    DECLARE_CLASS(APawn,AActor,0|CLASS_Abstract,XC_Engine)
};

//
// Editor entry point
//
class XC_ENGINE_API UXC_EditorLoader : public UObject
{
	public:
	DECLARE_CLASS(UXC_EditorLoader,UObject,0,XC_Engine)
};
