
inline FLOAT _Reciprocal( FLOAT F)
{
#if ASM
	FLOAT z;
	__asm
	{
		rcpss    xmm0, F           // x0: z ~= 1/x
		movss    xmm2, F           // x2: x
		movss    xmm1, xmm0        // x1: z ~= 1/x
		addss    xmm0, xmm0        // x0: 2z
		mulss    xmm1, xmm1        // x1: z^2
		mulss    xmm1, xmm2        // x1: xz^2
		subss    xmm0, xmm1        // x0: z' ~= 1/x to 0.000012%
		movss    z, xmm0          
    }
    return z;
#elif ASMLINUX
	//Higor: finally got this working on GCC 2.95
	FLOAT z;
	__asm__ __volatile__("rcpss    %0,%%xmm0\n"
			"movss    %0,%%xmm2\n"
			"movss    %%xmm0,%%xmm1\n"
			"addss    %%xmm0,%%xmm0\n"
			"mulss    %%xmm1,%%xmm1\n"
			"mulss    %%xmm2,%%xmm1\n"
			"subss    %%xmm1,%%xmm0\n" : : "m" (F) );
	__asm__ __volatile__("movss    %%xmm0,%0\n" : "=m" (z) );
	return z;
#else
	return 1.f/F; //ALL ABOARD THE SLOW TRAIN
#endif
}


#if ASM
inline void ASMTransformXY(const FCoords &Coords, const FVector &InVector, FVector &OutVector)
{
	// FCoords is a structure of 4 vectors: Origin, X, Y, Z
	//				 	  x  y  z
	// FVector	Origin;   0  4  8
	// FVector	XAxis;   12 16 20 <= z=0
	// FVector  YAxis;   24 28 32 <= z=0
	// FVector  ZAxis;   36 40 44 <= unused
	//
	//	task:	Temp = ( InVector - Coords.Origin );
	//			Outvector.X = (Temp | Coords.XAxis);
	//			Outvector.Y = (Temp | Coords.YAxis);
	//			Outvector.Z = (Temp);
	// Basically, move(X,Y,Z) and rotate(X,Y) a point
	// This is super fast
	__asm
	{
		mov     esi,[InVector]
		mov     edx,[Coords]     
		mov     edi,[OutVector]

		// get source
		fld     dword ptr [esi+0]
		fld     dword ptr [esi+4]
		fld     dword ptr [esi+8] // z y x
		fxch    st(2)     // xyz

		// subtract origin
		fsub    dword ptr [edx + 0]  // xyz
		fxch    st(1)  
		fsub	dword ptr [edx + 4]  // yxz
		fxch    st(2)
		fsub	dword ptr [edx + 8]  // zxy

		fstp	dword ptr [edi+8] // store z, xy
		// duplicate X for transforming
		fld		st(0)			// X X Y
		fmul	dword ptr [edx+12]	//Xx X Y
		fxch	st(1)			//X Xx Y
		fmul	dword ptr [edx+24]	//Xy Xx Y

		// duplicate Y for transforming
		fxch	st(2)			//Y Xx Xy
		fld		st(0)			//Y Y Xx Xy
		fmul	dword ptr [edx+28]	//Yy Y Xx Xy
		fxch	st(1)			//Y Yy Xx Xy
		fmul	dword ptr [edx+16]	//Yx Yy Xx Xy
		
		// sum results (Xx+Yx and Xy+Yy)
		faddp	st(2),st(0)		//Yy  Yx+Xx  Xy
		faddp	st(2),st(0)		//Yx+Xx  Yy+Xy

		// store
		fstp	dword ptr [edi+0]
		fstp	dword ptr [edi+4]
	}
}
#elif ASMLINUX
inline void ASMTransformXY(const FCoords &Coords, const FVector &InVector, FVector &OutVector)
{
	__asm__ __volatile__ ("
		# Get source.
		flds	0(%%esi);			# x
		flds	4(%%esi);			# y x
		flds	8(%%esi);			# z y x
		fxch	%%st(2);

		# Subtract origin.
		fsubs	0(%1);
		fxch	%%st(1);
		fsubs	4(%1);
		fxch	%%st(2);
		fsubs	8(%1);				# z x y

		fstps	8(%%edi);			# store z, xy
		# duplicate X for transforming
		fld		%%st(0); 			# X X Y
		fmuls	12(%1);				# Xx X Y
		fxch	%%st(1);			# X Xx Y
		fmuls	24(%1);				# Xy Xx Y

		# duplicate Y for transforming
		fxch	%%st(2);			# Y Xx Xy
		fld		%%st(0);			# Y Y Xx Xy
		fmuls	28(%1);				# Yy Y Xx Xy
		fxch	%%st(1);			# Y Yy Xx Xy
		fmuls	16(%1);				# Yx Yy Xx Xy

		# sum results (Xx+Yx and Xy+Yy)
		faddp	%%st(0),%%st(2);	# Yy  Yx+Xx  Xy
		faddp	%%st(0),%%st(2);	# Yx+Xx  Yy+Xy

		# store
		fstps	0(%%edi);
		fstps	4(%%edi);
	"
	:
	:	"S" (&InVector),
		"q" (&Coords),
		"D" (&OutVector)
	: "memory"
	);
}
#endif




inline FVector TransformPointByXY( const FCoords &Coords, const FVector& Point )
{
#if ASM
	FVector Temp;
	ASMTransformXY( Coords, Point, Temp);
	return Temp;
#elif ASMLINUX
	static FVector VTemp;
	ASMTransformXY( Coords, Point, VTemp);
	return VTemp;
#else
	FVector Temp = Point - Coords.Origin;
	return FVector(	Temp | Coords.XAxis, Temp | Coords.YAxis, Temp);
#endif
}

inline FLOAT _appInvSqrt( FLOAT F )
{
#if ASM
	const FLOAT fThree = 3.0f;
	const FLOAT fOneHalf = 0.5f;
	FLOAT temp;

	__asm
	{
		movss	xmm1,[F]
		rsqrtss	xmm0,xmm1			// 1/sqrt estimate (12 bits)

		// Newton-Raphson iteration (X1 = 0.5*X0*(3-(Y*X0)*X0))
		movss	xmm3,[fThree]
		movss	xmm2,xmm0
		mulss	xmm0,xmm1			// Y*X0
		mulss	xmm0,xmm2			// Y*X0*X0
		mulss	xmm2,[fOneHalf]		// 0.5*X0
		subss	xmm3,xmm0			// 3-Y*X0*X0
		mulss	xmm3,xmm2			// 0.5*X0*(3-Y*X0*X0)
		movss	[temp],xmm3
	}
	return temp;
#elif ASMLINUX
	//Higor: finally got this working on GCC 2.95
	const FLOAT fThree = 3.0f;
	const FLOAT fOneHalf = 0.5f;
	FLOAT temp;

	__asm__ __volatile__("movss    %0,%%xmm1\n"
			"rsqrtss    %%xmm1,%%xmm0\n"
			"movss    %1,%%xmm3\n"
			"movss    %%xmm0,%%xmm2\n"
			"mulss    %%xmm1,%%xmm0\n"
			"mulss    %%xmm2,%%xmm0\n"
			"mulss    %2,%%xmm2\n"
			"subss    %%xmm0,%%xmm3\n"
			"mulss    %%xmm2,%%xmm3\n" : : "m" (F), "m" (fThree), "m" (fOneHalf) );
	__asm__ __volatile__("movss    %%xmm3,%0\n" : "=m" (temp) );
	return temp;
#else
	return 1.0f / appSqrt( F);
#endif
}


inline FVector _UnsafeNormal( const FVector& V)
{
	const FLOAT Scale = _appInvSqrt(V.X*V.X+V.Y*V.Y+V.Z*V.Z);
	return FVector( V.X*Scale, V.Y*Scale, V.Z*Scale );
}

inline FVector _UnsafeNormal2D( const FVector& V)
{
	const FLOAT Scale = _appInvSqrt(V.X*V.X+V.Y*V.Y);
	return FVector( V.X*Scale, V.Y*Scale, 0.f);
}

inline FVector _UnsafeNormal2D( const FLOAT& X, const FLOAT& Y)
{
	const FLOAT Scale = _appInvSqrt(X*X+Y*Y);
	return FVector( X*Scale, Y*Scale, 0.f);
}
