/*=============================================================================
	Advanced travel manager functions
	By Higor, feel free to use this code on your project.
=============================================================================*/

#include "XC_Engine.h"

struct FStrPair
{
	INT HashNext;
	FString Key;
	FString Value;
};

UBOOL FXC_TravelManager::Exec( const TCHAR* Cmd, FOutputDevice& Ar )
{
	guard(FXC_TravelManager::Exec);
	if ( ParseCommand(&Cmd,TEXT("TRAVELINFO")) )
	{
		TArray<FStrPair>* Map = (TArray<FStrPair>*) (DWORD) &GetActiveLevel(Engine)->TravelInfo;
		debugf((EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("Map has %i entries"), Map->Num() );
		for ( int i=0 ; i<Map->Num() ; i++ )
			debugf((EName)XC_ENGINE_XC_Engine.GetIndex(), TEXT("Travel %s > %s"), *((*Map)(i).Key), *((*Map)(i).Value) );
	}
	else if ( ParseCommand(&Cmd,TEXT("TRAVELUPDATE")) )
		Poll( 0.f, true);
	else
		return 0;
	return 1;
	unguard;
}

struct FIteratorPList
{
	APlayerPawn* P;
	FIteratorPList* Next;
	FIteratorPList() {};
	FIteratorPList( APlayerPawn* nP, FIteratorPList* InNext)
	:	P(nP)
	,	Next(InNext)
	{};
};

void FXC_TravelManager::Poll( FLOAT DeltaTime, UBOOL bForceUpdate)
{
	if ( bNoPoll )
		return;
	guard( FXC_TravelManager::Poll);
	ULevel* Lev = GetActiveLevel( Engine);

	FMemMark Mark(GMem);
	FIteratorPList* BaseL = NULL;
	if ( Engine->Client && Engine->Client->Viewports.Num() && Engine->Client->Viewports(0) && Engine->Client->Viewports(0)->Actor )
		BaseL = new(GMem) FIteratorPList( Engine->Client->Viewports(0)->Actor, BaseL);
	if ( Lev && Lev->NetDriver && Lev->NetDriver->ClientConnections.Num() ) //Clients
	{
		for ( INT ip=0 ; ip < Lev->NetDriver->ClientConnections.Num() ; ip++ )
			if ( Lev->NetDriver->ClientConnections(ip) && Lev->NetDriver->ClientConnections(ip)->Actor )
				BaseL = new(GMem) FIteratorPList( Lev->NetDriver->ClientConnections(ip)->Actor, BaseL);
	}

	for ( FIteratorPList* Link=BaseL ; Link ; Link=Link->Next )
		FixUpPlayerId( Link->P);
	if ( bAutoMode || bForceUpdate ) //Timed every 1-2 seconds to scan players and update their stuff
	{
		TimerCounter -= DeltaTime;
		if ( bForceUpdate || TimerCounter < 0 )
		{
			for ( Link=BaseL ; Link ; Link=Link->Next )
				GenerateTravelInfo( Link->P);
			TimerCounter = 2.0;
		}
	}
	Mark.Pop();
	unguard;
}


void FXC_TravelManager::FixUpPlayerId( APlayerPawn* P)
{
	guard(FXC_TravelManager::FixUpPlayerId);
	check( P);
	if ( P->PlayerReplicationInfo )
	{
		INT i;
		for ( i=0 ; i<PlayerMap.Num() ; i++ )
			if ( PlayerMap(i).PlayerID == P->PlayerReplicationInfo->PlayerID ) //Found
			{
				if ( appStrcmp( *(P->PlayerReplicationInfo->PlayerName), *PlayerMap(i).PlayerName) ) //Name mismatch
				{
					for ( INT j=0 ; j<PlayerMap.Num() ; j++ )
						if ( !appStrcmp( *(P->PlayerReplicationInfo->PlayerName), *PlayerMap(j).PlayerName) ) //Name match
						{
							P->PlayerReplicationInfo->PlayerName = PlayerMap(j).PlayerName; //Revert name
							return;
						}
					GetActiveLevel( Engine)->TravelInfo.Remove( *PlayerMap(i).PlayerName );
					PlayerMap(i).PlayerName = P->PlayerReplicationInfo->PlayerName;
					GetActiveLevel( Engine)->TravelInfo.Set( *PlayerMap(i).PlayerName, *PlayerMap(i).TravelList);
				}
				return;
			}
		//Not in map, let's find
		for ( i=0 ; i<PlayerMap.Num() ; i++ )
			if ( PlayerMap(i).PlayerName == P->PlayerReplicationInfo->PlayerName ) //Found
			{
				PlayerMap(i).PlayerID = P->PlayerReplicationInfo->PlayerID; //Attach
				return;
			}

		//Not in map, let's add
		PlayerMap.AddZeroed();
		PlayerMap.Last().PlayerID = P->PlayerReplicationInfo->PlayerID;
		PlayerMap.Last().PlayerName = P->PlayerReplicationInfo->PlayerName;
	}
	unguard;
}

void FXC_TravelManager::GenerateTravelInfo( APlayerPawn* P)
{
	guard(FXC_TravelManager::GenerateTravelInfo);
	FString STR = FString();
	FString NewLine = FString::Chr(0x0D) + FString::Chr(0x0A);
	TCHAR Temp[256] = TEXT("");
	for ( AActor* A = P ; A ; A = A->Inventory )
	{
		if ( A->bTravel )
		{
			INT Idx = TravelClassMap.RegisterClass( A->GetClass() );
			STR += FString::Printf(TEXT("Class=%s.%s Name=%s%s{%s"), A->GetClass()->GetOuter()->GetName(), A->GetClass()->GetName(), A->GetName(), *NewLine, *NewLine );
			while ( Idx >= 0 )
			{
				for ( int i=0 ; i<TravelClassMap(Idx).Properties.Num() ; i++ )
				{
					UProperty* Prop = TravelClassMap(Idx).Properties(i);
					check( Prop);

					if ( Prop->Matches( A, A->GetClass()->GetDefaultActor(), 0) ) //Same as default, disregard ||TREAT AS ARRAY (3RD PARAMETER)
						continue;
						//Should be treating this as array!!!
					guard(TEST2);
					Prop->ExportText( 0, Temp, (BYTE*)A,(BYTE*)A, PPF_Localized );
					unguard;
					if ( Prop->IsA( UObjectProperty::StaticClass() ) && !Prop->Matches( A, 0, 0) )
					{
						UObject** TestObj = (UObject**)  ( DWORD(A) + Prop->Offset );
						STR += FString::Printf( TEXT("%s=%s%s%s%s"), Prop->GetName(), (*TestObj)->GetName(), *NewLine, Temp, *NewLine);
					}
					else
						STR += FString::Printf(  TEXT("%s=%s%s"), Prop->GetName(), Temp, *NewLine);
				}
				Idx = TravelClassMap(Idx).ParentIdx;
			}
			STR += FString::Printf( TEXT("}%s"), *NewLine);
		}
	}

	INT i;
	for ( i=0 ; i<PlayerMap.Num() ; i++ )
		if ( PlayerMap(i).PlayerID == P->PlayerReplicationInfo->PlayerID ) //Found
		{
			PlayerMap(i).TravelList = STR;
			break;
		}
	GetActiveLevel( Engine)->TravelInfo.Set( *P->PlayerReplicationInfo->PlayerName, *STR);
	unguard;
}

void FXC_TravelManager::PostMapChange( ULevel* NewLevel, UBOOL bClearMap)
{
	guard( FXC_TravelManager::PostMapChange);
	//Empty the map... let's not support cross map remembering yet..
	if ( bClearMap && (PlayerMap.Num() > 0) )
		PlayerMap.Empty();
	TimerCounter = 2;

	bNoPoll = 0;
	if ( !NewLevel || !NewLevel->IsServer() || !NewLevel->Actors.Num() || !NewLevel->NetDriver )
		bNoPoll = 1;

	if ( !bNoPoll && !bClearMap ) //Set new level's travel list for multi travel
	{
		for ( INT i=0 ; i<PlayerMap.Num() ; i++ )
			NewLevel->TravelInfo.Set( *PlayerMap(i).PlayerName, *PlayerMap(i).TravelList);
	}
	unguard;
}

INT FTravelClassMap::RegisterClass( UClass* InClass) //Return index in map
{
	guard(FTravelClassMap::RegisterClass);
	if ( !InClass->IsChildOf( AActor::StaticClass()) )
		return -1;
	if ( !ArrayNum )
	{
		AddZeroed();
		(*this)(0).ParentIdx = -1;
		(*this)(0).Class = AActor::StaticClass();
		if ( InClass == AActor::StaticClass() )
			return 0;
	}
	//Find direct subclasses, register whatever's necessary
	UClass* Parent = AActor::StaticClass();
	INT ParentIdx = 0;
	while ( Parent )
	{
		UBOOL Found = 0;
		UClass* CurClass = InClass;
		if ( CurClass == Parent ) //Cur has been added!
			break;
		while ( CurClass->GetSuperClass() != Parent )
			CurClass = CurClass->GetSuperClass();
		//This class's parent is obviously on this list, attempt to find using parent's children array
		for ( int i=0 ; i< (*this)(ParentIdx).ChildrenIdx.Num() ; i++ )
		{
			if ( (*this)( (*this)(ParentIdx).ChildrenIdx(i) ).Class == CurClass ) //This class is listed on parent's children, do not register
			{
				Found = 1;
				ParentIdx = (*this)(ParentIdx).ChildrenIdx(i);
				break;
			}
		}
		//Register this class if necessary
		if ( !Found )
		{
			AddZeroed();
			Last().Class = CurClass;
			Last().Idx = ArrayNum - 1;
			Last().ParentIdx = ParentIdx;
			(*this)(ParentIdx).ChildrenIdx.Add();
			(*this)(ParentIdx).ChildrenIdx.Last() = Last().Idx; //Add the new child IDX (this) onto the parent
			
			UField* Field = CurClass->Children;
			while ( Field ) //Find Travel properties
			{
				if ( Field->IsA( UProperty::StaticClass()) && (((UProperty*)Field)->PropertyFlags & CPF_Travel) ) //Travel property, list it
				{
					Last().Properties.AddZeroed();
					Last().Properties.Last() = (UProperty*) Field;
				}
				Field = Field->Next;
			}
			ParentIdx = Last().Idx;
		}
		Parent = CurClass;
	}
	return ParentIdx;
	unguard;
}


/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
